/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */


/*
 * Probability.hpp
 *
 *  Created on: 2010/12/20
 *      Author: HARK Development Team
 */

#ifndef PROBABILITY_HPP_
#define PROBABILITY_HPP_

#include <stdexcept>

namespace NSHarkBinaural
{

/**
 * 確率を表すベースクラス.
 */
template <typename T>
class ProbabilityBase
{
public:
    /**
     * 指定された確率を保持したインスタンス生成する.
     * 指定された確率の値が、0以上1以下に含まれない場合、
     * std::out_of_range例外が発生する。
     * @param prob 確率
     */
    ProbabilityBase(T prob = T())
        : prob(prob)
    {
        CheckRange(prob);
    }

    /**
     * 指定された確率を設定する.
     * 指定された確率の値が、0以上1以下に含まれない場合、
     * std::out_of_range例外が発生する。
     * @param prob 確率
     */
    void Set(T prob)
    {
        CheckRange(prob);
        this->prob = prob;
    }

    /**
     * インスタンスが保持する確率を得る.
     * @return 確率
     */
    T Get() const
    {
        return this->prob;
    }

    /**
     * Tへの変換.
     * @return 確率
     */
    operator T () const
    {
        return this->prob;
    }
private:
    /**
     * 確率の値の範囲をチェックする.
     * 0以上1以下でない場合、std::out_of_range例外を発生させる。
     * @param prob 確率
     */
    void CheckRange(T prob)
    {
        if (prob < 0) throw std::out_of_range("probability is less than 0");
        if (prob > 1) throw std::out_of_range("probability is more than 1");
    }

    /**
     * 確率.
     */
    T prob;
};

typedef ProbabilityBase<double> Probability;

}

#endif /* PROBABILITY_HPP_ */
