/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */


/*
 * Clustering.cpp
 *
 *  Created on: 2010/12/16
 *      Author: HARK Development Team
 */

#include <vector>
#include <stdexcept>
#include <boost/shared_ptr.hpp>
#include <Clustering.hpp>
#include <Cluster.hpp>
#include <Signal.hpp>
#include <Band.hpp>
#include <Logger.hpp>

using namespace std;
using namespace boost;

namespace NSHarkBinaural
{

Clustering::ClusterVectorPtr
Clustering::MakeInitClusters(Peak::PeakVectorPtr peak)
{
    static Logger& logger = Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME);

    if (peak.get() == 0)
    {
        LogWarning(logger, "peak is null");
        return ClusterVectorPtr();
    }

    ClusterVectorPtr result(new ClusterVector());

    Peak::PeakVector& peak_ = *peak;
    ClusterVector& clusters = *result;

    ClusterPtr cluster;

    clusters.reserve(peak_.size());

    for (size_t i = 0; i < peak_.size(); ++i)
    {
        /*
        if (peak_[i].power > minDB)
        {
          LogFinest(logger, format("adding a cluster, No %d") % i);
          cluster = ClusterPtr(new Cluster(bi.CreateBand(i), peak_[i].power));
          clusters.push_back(cluster);
        }
        */
        LogFinest(logger, format("adding a cluster, No %d, band %d") % i % peak_[i].band.Index());
        cluster = ClusterPtr(new Cluster(peak_[i].band, peak_[i].power));
        clusters.push_back(cluster);
    }
    LogFine(logger, format("made %d clusters") % clusters.size());

    return result;
}


Clustering::CostAndIndex
Clustering::GetClosestDst(ClusterVectorPtr clusters)
{
    static Logger& logger = Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME);

    CostAndIndex cai = { numeric_limits<double>::max(), -1, -1 };
    if (clusters.get() == 0)
    {
        LogWarning(logger, "clusters is null");
        return cai;
    }

    ClusterVector& clusters_ = *clusters;
    double cost;

    if (clusters_.size() == 0) return cai;

    LogFine(logger, "calculating cost and index");
    for (size_t i = 0; i < clusters_.size() - 1; ++i)
    {
        if (clusters_[i].get() == 0) LogWarning(logger, format("clusters_[%d] is null") % i);
        if (clusters_[i + 1].get() == 0) LogWarning(logger, format("clusters_[%d] is null") % (i + 1));
        try
        {
            cost = clusters_.at(i)->CostF(*clusters_.at(i + 1));
        }
        catch (std::out_of_range& e)
        {
            LogWarning(logger, e.what());
            throw e;
        }
        if (cost < cai.cost)
        {
            cai.cost = cost;
            cai.index1 = i;
            cai.index2 = i + 1;
        }
    }
    return cai;
}

Clustering::ClusterVectorPtr
Clustering::Execute(Peak::PeakVectorPtr peak, double maxCost)
{
    static Logger& logger = Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME);

    ClusterVectorPtr clusters = MakeInitClusters(peak);

    CostAndIndex cai = GetClosestDst(clusters);

    LogFine(logger, format("updating clusters and calculating closest dst: init cost = %f") % cai.cost);
    while (cai.cost < maxCost && clusters->size() > 1)
    {
        clusters->at(cai.index1)->Merge(*clusters->at(cai.index2));
        clusters->erase(clusters->begin() + cai.index2);
        cai = GetClosestDst(clusters);
    }

    LogFine(logger, "creating result variable");
    // メモリ節約のため、空の配列を生成して、中身を交換する
    ClusterVectorPtr result(new ClusterVector());
    if (result.get() != 0) result->swap(*clusters);
    return result;
}

Clustering::BandVectorPtr
Clustering::GetProxy(ClusterVectorPtr clusters, Unspectrum::ptr_t bgnoise)
{
    static Logger& logger = Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME);

    // 引数のチェック
    if (clusters.get() == 0 || bgnoise.get() == 0)
    {
        LogWarning(logger, format("clusters or bgnoise is null"));
        return BandVectorPtr();
    }

    // 使いやすくするための引数の参照はずし
    ClusterVector& clusters_ = *clusters;
    Unspectrum& bgnoise_ = *bgnoise;

    // 結果を格納するリストの生成
    BandVectorPtr result(new BandVector());
    BandVector& result_ = *result;
    result_.reserve(clusters_.size());

    // 1つのクラスタ内での最大値
    double max;
    // 最大値に対応する要素
    vector<BandAndPower>::const_iterator maxi;

    ClusterPtr cluster;
    // クラスタ群を走査する
    for (ClusterVector::iterator it = clusters_.begin(); it != clusters_.end(); ++it)
    {
        cluster = *it;
        if (cluster.get() == 0)
        {
            // このif文は実行されないはず。一応ログの処理は記述しておく。
            LogSevere(logger, format("cluster in cluster vector is null"));
            // 引数のリストのインデックスと、戻り値のリストのインデックスを合わせるため、
            // ダミーをプッシュしておく。
            result_.push_back(Band());
        }
        else if (cluster->BandAndPowerList().size() == 0)
        {
            // クラスタが1つも周波数バンドを保持していない状態は異常である。
            LogSevere(logger, format("band & power list size is zero"));
            // 上と同様にダミーをプッシュしておく。
            result_.push_back(Band());
        }
        else
        {
            // 使いやすくするため名前をつける。
            const vector<BandAndPower>& bplist = cluster->BandAndPowerList();

            // クラスタが保持する周波数バンドの中で、最大のパワーをもつものを代表値とする。
            maxi = bplist.begin();
            max = bplist.front().power - bgnoise_.At(bplist.front().band.Index());
            LogFine(logger, format("initialize max : %f") % max);

            // クラスタ中の周波数バンドを走査する。
            for (vector<BandAndPower>::const_iterator bp = bplist.begin() + 1; bp != bplist.end(); ++bp)
            {
                if (bp->power - bgnoise_.At(bp->band.Index()) > max)
                {
                    LogFinest(logger, format("bp->power = %f, bgnoise_.At(%d) = %f")
                              % bp->power % bp->band.Index() % bgnoise_.At(bp->band.Index()));
                    max = bp->power - bgnoise_.At(bp->band.Index());
                    maxi = bp;
                }
            }

            // 見つけた最大パワーの周波数バンドを代表値リストに追加する。
            result_.push_back(maxi->band);
        }
    }

    return result;
}

}
